use proc_macro2::{Ident, Span, TokenStream};
use serde::{de::DeserializeOwned, Serialize};
use syn::{punctuated::Punctuated, token::Comma, FnArg, Pat};

struct RecordableCommand<T: Serialize + DeserializeOwned, R> {
    args: T,
    func: dyn Fn(T) -> R,
}

/// Takes a function and produces a serializable struct of its args.
pub fn arg_struct(mut function: syn::ItemFn) -> TokenStream {
    let struct_name = Ident::new(&format!("{}_args", function.sig.ident), Span::call_site());
    let wrapped_fn_name = Ident::new(&format!("{}_fn", function.sig.ident), Span::call_site());

    let fn_name = &function.sig.ident;
    let fn_generics = &function.sig.generics;
    let fn_inputs = &function.sig.inputs;
    let fn_args = &function
        .sig
        .inputs
        .iter()
        .map(|arg| match arg {
            FnArg::Receiver(_) => {
                panic!("Can't accept Receiver arg")
            }
            FnArg::Typed(p) => p.pat.clone(),
        })
        .collect::<Punctuated<Box<Pat>, Comma>>();
    let fn_retval = &function.sig.output;
    let fn_body = &function.block;

    quote::quote! {
        pub async fn #fn_name #fn_generics(#fn_inputs) -> #fn_retval {
            #wrapped_fn_name(#fn_args).await
        }

        pub async fn #wrapped_fn_name #fn_generics(#fn_inputs) -> #fn_retval {
            #fn_body
        }

    }
}

/*

#[poise]
#[wrapper]
pub async fn command(...args) {
    ...block
}

 ... becomes ...

#[poise]
fn command(...args) {
    command_fn(
        ...args
    )
}

struct RecordableCommand<T> {
    args: T
    func: Func<T>
}

impl Execute<T> for RecordableCommand<T> {
    fn execute() {
        // Unpack self.args into self.func
    }
}

struct command_args {
    ...args
}

fn command_fn(...args) {
    ...block
}

 */
